/*
 * decode.c
 *
 * Copyright (C) Aaron Holtzman <aholtzma@ess.engr.uvic.ca> - Nov 1999
 *
 * Decodes an MPEG-2 video stream.
 *
 * This file is part of mpeg2dec, a free MPEG-2 video stream decoder.
 *	
 * mpeg2dec is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 * 
 * mpeg2dec is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with GNU Make; see the file COPYING. If not, write to
 * the Free Software Foundation, 
 *
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>

#include "xine.h"
#include "mpeg2.h"
#include "mpeg2_internal.h"
#include "monitor.h"

#ifdef __i386__
#include "mmx.h"
#include "sse.h"
#endif

extern uint32_t xine_debug;

//this is where we keep the state of the decoder
picture_t picture;

//global config struct
mpeg2_config_t config;

extern vo_driver_t *gVideoDriver;

//
//the current start code chunk we are working on
//we max out at 65536 bytes as that is the largest
//slice we will see in MP@ML streams.
// (we make no pretenses ofdecoding anything more than that)
static uint8_t chunk_buffer[224 * 1024 + 4];
static uint32_t shift = 0;

static uint32_t is_sequence_needed     = 1;

static int output_width                = 0;
static int output_height               = 0;
static int output_rate                 = 0;
static int output_ratio                = 0;

// sync stuff
static int bSkipFrames                 = 0;
static int bSkipBFrames                = 0;
static int bSkipSlices                 = 0;

static uint32_t given_video_pts;

int bMMX = 0;

int nVideoHours = 0;
int nVideoMins  = 0;
int nVideoSecs  = 0;

void mpeg2_init (void)
{

    int mm_supp;

    config.flags = 0;

#ifdef __i386__

    mm_supp = mm_support ();

    xprintf (VERBOSE, "mm_support : %d\n", mm_supp);

    if (mm_supp>0) {
      bMMX = 1;
      config.flags |= MPEG2_MMX_ENABLE;
    }

    if (mm_supp==9) 
      config.flags |= MPEG2_SSE_ENABLE;

#endif


    /* intialize the decoder state */
    shift = 0;
    is_sequence_needed = 1;

    header_state_init (&picture);
    idct_init ();
    motion_comp_init ();
    picture.forward_reference_frame_buf = NULL;
    picture.backward_reference_frame_buf = NULL;
    picture.throwaway_frame_buf = NULL;
}

void decode_free_image_buffers () {

    if (picture.forward_reference_frame_buf) {
      vo_free_image_buffer (picture.forward_reference_frame_buf);
      picture.forward_reference_frame_buf = NULL;
    }

    if (picture.backward_reference_frame_buf) {
      vo_free_image_buffer (picture.backward_reference_frame_buf);
      picture.backward_reference_frame_buf = NULL;
    }

    if (picture.throwaway_frame_buf) {
      vo_free_image_buffer (picture.throwaway_frame_buf);
      picture.throwaway_frame_buf = NULL;
    }

    picture.current_frame_buf = NULL;

    picture.forward_reference_frame[0] = NULL;
    picture.forward_reference_frame[1] = NULL;
    picture.forward_reference_frame[2] = NULL;
    
    picture.backward_reference_frame[0]  = NULL;
    picture.backward_reference_frame[1]  = NULL;
    picture.backward_reference_frame[2]  = NULL;
    
    picture.current_frame[0] = NULL;
    picture.current_frame[1] = NULL;
    picture.current_frame[2] = NULL;
    
    picture.throwaway_frame[0]  = NULL;
    picture.throwaway_frame[1]  = NULL;
    picture.throwaway_frame[2]  = NULL;

}

void mpeg2_reset (vo_driver_t *video_out) {
  
    shift = 0;
    is_sequence_needed    = 0; 
    picture.picture_coding_type = 0;

    /*
     * free buffers
     */

    decode_free_image_buffers ();

    bSkipFrames                 = 0;
    bSkipBFrames                = 0;
    bSkipSlices                 = 1;

    /*
    header_state_init (&picture);
    idct_init ();
    motion_comp_init ();
    */

    xprintf (VERBOSE|MPEG, "mpeg2_reset done\n");
}

static void decode_reorder_frames ()
{
    if (picture.picture_coding_type != B_TYPE) {

        if (picture.forward_reference_frame_buf)
          vo_free_image_buffer (picture.forward_reference_frame_buf);

	/*
	 * make the backward reference frame the new forward reference frame
	 */

	picture.forward_reference_frame[0] =
	    picture.backward_reference_frame[0];
	picture.forward_reference_frame[1] =
	    picture.backward_reference_frame[1];
	picture.forward_reference_frame[2] =
	    picture.backward_reference_frame[2];
	picture.forward_reference_frame_buf = 
	    picture.backward_reference_frame_buf;

	/*
	 * allocate new backward reference frame
	 */


	picture.backward_reference_frame_buf = vo_alloc_image_buffer();
	picture.backward_reference_frame[0]  = picture.backward_reference_frame_buf->mem[0];
	picture.backward_reference_frame[1]  = picture.backward_reference_frame_buf->mem[1];
	picture.backward_reference_frame[2]  = picture.backward_reference_frame_buf->mem[2];
	picture.backward_reference_frame_buf->PTS       = 0;
	picture.backward_reference_frame_buf->bFrameBad = 1;

	/*
	 * make it the current frame
	 */

	picture.current_frame[0] = picture.backward_reference_frame[0];
	picture.current_frame[1] = picture.backward_reference_frame[1];
	picture.current_frame[2] = picture.backward_reference_frame[2];
	picture.current_frame_buf = picture.backward_reference_frame_buf;

    } else {

	/*
	 * allocate new throwaway frame
	 */


	picture.throwaway_frame_buf = vo_alloc_image_buffer();
	picture.throwaway_frame[0]  = picture.throwaway_frame_buf->mem[0];
	picture.throwaway_frame[1]  = picture.throwaway_frame_buf->mem[1];
	picture.throwaway_frame[2]  = picture.throwaway_frame_buf->mem[2];
	picture.throwaway_frame_buf->PTS       = 0;
	picture.throwaway_frame_buf->bFrameBad = 1;

	/*
	 * make it the current frame
	 */

	picture.current_frame[0] = picture.throwaway_frame[0];
	picture.current_frame[1] = picture.throwaway_frame[1];
	picture.current_frame[2] = picture.throwaway_frame[2];
	picture.current_frame_buf= picture.throwaway_frame_buf;

    }

}


static int parse_chunk (int code, uint8_t * buffer, uint32_t pts)
{
    int is_frame_done = 0;


    /* printf ("mpeg2dec : parse_chunk, code=%02x\n",code); */

    if (is_sequence_needed && code != 0xb3)	/* b3 = sequence_header_code */
	return 0;

    /* stats_header (code, buffer); */


    switch (code) {
    case 0x00:	/* picture_start_code */

        /*
	  if (picture.picture_coding_type==D_TYPE)
	  printf ("D-Frame\n");
	*/

        if (picture.picture_coding_type==B_TYPE) {
	  
	  /*
	  if (picture.mpeg1) {
	    int slice_num = 1;
	    int num_slices = picture.coded_picture_height / 16;
	    while (slice_num < num_slices) {
	      gVideoDriver->notify_frame_portion_ready (picture.throwaway_frame_buf,slice_num);
	      slice_num++;
	    }
	  }
	  */
	  
	  vo_queue_frame(picture.throwaway_frame_buf);
	  vo_free_image_buffer (picture.throwaway_frame_buf);
	  picture.throwaway_frame_buf = NULL;

        } else {
	  
	    if (picture.forward_reference_frame_buf) {
	      /*
		if (picture.mpeg1) {
		    int slice_num = 1;
		    int num_slices = picture.coded_picture_height / 16;
		    while (slice_num < num_slices) {
			video_out->notify_frame_portion_ready (picture.forward_reference_frame_buf,
							       slice_num); 
			slice_num++;
		    }
		}
		*/

	        vo_queue_frame(picture.forward_reference_frame_buf);
	    }
	}


	header_process_picture_header (&picture, buffer);

	decode_reorder_frames ();

	
	if (given_video_pts) {
	    picture.current_frame_buf->PTS = given_video_pts;
	    /*
	      printf ("pts for image of type %d given : %d\n",
	      picture.picture_coding_type, pts);
	    */
	    given_video_pts = 0;
	}
	

	picture.current_frame_buf->nType = picture.picture_coding_type;

	switch (picture.picture_coding_type) {
	case B_TYPE:
	    if (bSkipFrames || bSkipBFrames) {
		bSkipSlices = 1;
		bSkipFrames = 0;
		bSkipBFrames = 0;
		return 0;
	    }
	    if (!picture.forward_reference_frame_buf
		|| !picture.backward_reference_frame_buf
		|| picture.forward_reference_frame_buf->bFrameBad 
		|| picture.backward_reference_frame_buf->bFrameBad) {
		bSkipSlices = 1;
		return 0;
	    }
	    bSkipSlices = 0;
	    break;
	case P_TYPE:
	    if (bSkipFrames) {
		bSkipSlices = 1;
		bSkipFrames = 0;
		return 0;
	    }
	    if (!picture.forward_reference_frame_buf
		|| picture.forward_reference_frame_buf->bFrameBad) {
		bSkipSlices = 1;
		return 0;
	    }
	    bSkipSlices = 0;
	    break;
	case I_TYPE:
	    if (bSkipFrames) {
		bSkipFrames = 0;
		bSkipSlices = 1;
		return 0;
	    }
	    bSkipSlices = 0;
	    break;
	}

	break;

    case 0xb3:	/* sequence_header_code */
	header_process_sequence_header (&picture, buffer);
	is_sequence_needed = 0;

	if (( picture.coded_picture_width != output_width) 
	    || (picture.coded_picture_height != output_height)
	    || (picture.aspect_ratio_information != output_ratio)
	    || (picture.frame_rate_code != output_rate)) {

	  decode_free_image_buffers ();

	  vo_set_size(picture.coded_picture_width,
		      picture.coded_picture_height,
		      picture.aspect_ratio_information,
		      picture.frame_rate_code);

	  output_width  = picture.coded_picture_width;
	  output_height = picture.coded_picture_height;
	  output_rate   = picture.frame_rate_code;
	  output_ratio  = picture.aspect_ratio_information;
	}
	break;

    case 0xb5:	/* extension_start_code */
	header_process_extension (&picture, buffer);
	break;

    case 0xb8:

      nVideoHours = (buffer[0]&0x7c)>>2;
      nVideoMins  = ((buffer[0]&0x03)<<4)|((buffer[1]&0xF0)>>4);
      nVideoSecs  = ((buffer[1]&0x07)<<3)|((buffer[2]&0xe0)>>5);

      /*
      printf ("GOP %02d:%02d:%02d (%02d)\n",
              (buffer[0]&0x7c)>>2,
              ((buffer[0]&0x03)<<4)|((buffer[1]&0xF0)>>4),
              ((buffer[1]&0x07)<<3)|((buffer[2]&0xe0)>>5),
              ((buffer[2]&0x1F)<<1)|((buffer[3]&0x80)>>7));
	      */
      break;
    default:
	if ((code < 0xb0) && code) {

	  if (!bSkipSlices) {

	    is_frame_done = slice_process (&picture, code, buffer);

	    if (is_frame_done)
	      picture.current_frame_buf->bFrameBad = 0;

	  } 

	}
    }

    return is_frame_done;
}

uint8_t *mpeg2_decode_data (uint8_t * current, uint8_t * end, uint32_t pts)
{
    static uint8_t code = 0xff;
    //static uint8_t chunk_buffer[65536];
    static uint8_t *chunk_ptr = chunk_buffer;
    //static uint32_t shift = 0;

    uint8_t byte;
    int frame_done = 0;
    
    /* printf ("mpeg2_decode_data, pts = %d\n",pts);   */

    if (pts)
      given_video_pts = pts;

    while (current != end) {
	while (1) {
	    byte = *current++;
	    if (shift == 0x00000100)
		break;
	    *chunk_ptr++ = byte;
	    shift = (shift | byte) << 8;

	    if (current == end) {
#ifdef __i386__
	        if (bMMX) 
 		    emms();
#endif

		return current;
	    }
	}

	/* found start_code following chunk */

	frame_done = parse_chunk (code, chunk_buffer, pts) ;

	/* done with header or slice, prepare for next one */

	code = byte;
	chunk_ptr = chunk_buffer;
	shift = 0xffffff00;

	if (frame_done) {
#ifdef __i386__
	  if (bMMX) 
	    emms();
#endif
	  return current;
	}

    }

#ifdef __i386__
    if (bMMX) 
      emms();
#endif
    return current;
}

void mpeg2_set_skip_frame () {
  
  bSkipFrames = 1;
  bSkipBFrames = 0;

}

void mpeg2_set_skip_bframe () {
  
  bSkipBFrames = 1;
  
}

void mpeg2_set_skip_slices () {
  bSkipSlices  = 1;
}
